#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>                                                       
#include <unistd.h>
#include <errno.h>  
#include <string.h>

#include <UDCF.h>
#include <plot.h>
#include "extern.h"
#include "bool.h"
#include "../config.h"

static bool verbose;
static bool useMixes;

static const double plotWidth = 0.75;
static const double plotHeight = 0.6;

#define ILLEGAL "illegal"
const static char* formats[] = {"ps", "ai", "fig", "pnm", ILLEGAL};
const static char* papers[] = {"a4", "a3", "a2", "a1", "a0", 
                               "letter", ILLEGAL};

const char* GetEntry(const char* search, const char* searchSpace[])
{
 int i = 0;

 /* ---- Check all entries in 'searchSpace' ---- */
 while ((strcmp(searchSpace[i], search) != 0) && 
        (strcmp(searchSpace[i], ILLEGAL) != 0)) i++;

 /* ---- If 'search' was found return it. If not then use first array element ---- */
 if (strcmp(searchSpace[i], ILLEGAL) != 0)
  return searchSpace[i];
 else
  return searchSpace[0];

}

void ConnectToFile(const char* filename, int openFlags, int connectTo)
{
 int fileNo;

 /* ---- Open specified filename ---- */
 if ((fileNo = open(filename, 
                    openFlags, 
                    S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) == -1) {

  /* ---- Give an error and exit ---- */
  fprintf(stderr, "Can't open %s (%s)!\n", filename, strerror(errno));
  exit(EXIT_FAILURE);

 }            

 /* ---- Try to close 'connectTo' ---- */
 if (close(connectTo) == -1) {

  /* ---- Give an error and exit ---- */
  fprintf(stderr, "Can't close channel (%s)!\n", strerror(errno));
  exit(EXIT_FAILURE);

 }
 
 /* ---- Connect 'connectTo' to file 'fileNo' ---- */
 if (dup(fileNo) == -1) {

  /* ---- Give an error and exit ---- */
  fprintf(stderr, "Can't connect to %s (%s)!\n", filename, strerror(errno));
  exit(EXIT_FAILURE);

 }

 /* ---- Close 'fileNo' since it is connected to 'connectTo' now ---- */
 if (close(fileNo) == -1) {

  /* ---- Give an error and exit ---- */
  fprintf(stderr, "Can't close file after dup (%s)!\n", strerror(errno));

 }

}

void PrintHelp(const char* prgname)
{
#ifdef HAVE_GETOPTLONG
 /* ---- Print help message ---- */
 fprintf(stderr, "Usage: %s [OPTION(s)]\n", prgname);
 if (verbose)
  fprintf(stderr, "$Id: udcfplot.c,v 1.14 2001/08/29 09:47:22 streit Exp $\n");
 fprintf(stderr, 
         "-f, --format [ps|ai|fig|pnm]    Select format to export. Default: '%s'\n",
         formats[0]); 
 fprintf(stderr, 
         "-p, --paper [a{0-4}|letter]     Select paper type. Default: '%s'\n",
         papers[0]); 
 fprintf(stderr, 
         "-g, --group number              Select group of repetitive dives. ");
 fprintf(stderr,"Default: '0'\n");
 fprintf(stderr, 
         "-d, --dive  number              Select dive in actual group of ");
 fprintf(stderr, "repetitive dives. Default: '0'\n");
 fprintf(stderr, 
         "-r, --ranges                    Give valid ranges of groups and dives\n"); 
 fprintf(stderr, "-h, --help                      Print a help text\n");
 fprintf(stderr, "-m, --mixes                     Show infos about mixes used\n");
 fprintf(stderr, "-v, --verbose                   Explain what is going on\n"); 
 fprintf(stderr, "-i, --input filename            Read UDCF-data from 'filename'\n");
 fprintf(stderr, "-o, --output filename           Write data to 'filename'\n");
 fprintf(stderr, "\n");
#else
 /* ---- Print help message ---- */
 fprintf(stderr, "Usage: %s [OPTION(s)]\n", prgname);
 if (verbose)
  fprintf(stderr, "$Id: udcfplot.c,v 1.14 2001/08/29 09:47:22 streit Exp $\n");
 fprintf(stderr, 
         "-f [ps|ai|fig|pnm] Select format to export. Default: '%s'\n",
         formats[0]); 
 fprintf(stderr, 
         "-p [a{0-4}|letter] Select paper type. Default: '%s'\n",
         papers[0]); 
 fprintf(stderr, 
         "-g number, Select group of repetitive dives. Default: '0'\n");
 fprintf(stderr, 
         "-d number, Select dive in actual group of repetitive dives. Default: '0'\n");
 fprintf(stderr, "-h, Print a help text\n");
 fprintf(stderr, "-m, Show infos about mixes used\n");
 fprintf(stderr, "-v, Explain what is going on\n");
 fprintf(stderr, "-i, Read UDCF-data from 'filename'\n");
 fprintf(stderr, "-o, Write data to 'filename'\n");
 fprintf(stderr, "\n");
#endif 

 /* ---- Terminate program successfully ---- */
 exit(EXIT_SUCCESS);
 
}

void PlotInfo(UDCFDive* tdDive, Multigrapher* mg)
{
 const int buffSize = 1024;
 char buffer[buffSize];
 long hours,mins;
 long i;

 /* ---- Check for valid dive-struct ---- */
 if (tdDive == (UDCFDive *) NULL) 
#ifndef NDEBUG
 /* ---- Give an error and exit ---- */
 {
  fprintf(stderr, "Can't plot dive information!\n");
  exit(EXIT_FAILURE);
 }
#else
 /* ---- Simply return if invalid ---- */
 return;
#endif

 /* ---- Render time and depth ---- */
 for (i = 0; i < buffSize; i++) buffer[i]='\0';
 snprintf(buffer,
          buffSize - 1,
          "Date&Time: %4.4d/%2.2d/%2.2d %2.2d:%2.2d",
          tdDive -> year,
          tdDive -> month,
          tdDive -> day,
          tdDive -> hour,
          tdDive -> minute);

 /* ---- Plot time and date ---- */
 draw_label(mg, buffer, 0.0, 0.9);
 
 /* ---- Render pressure ---- */
 for (i = 0; i < buffSize; i++) buffer[i]='\0';
 snprintf(buffer,
          buffSize - 1,
          "Pressure (Mix: %s): %4.1f - %4.1f Bar", 
          tdDive -> mixList[0].mixName,
          tdDive -> mixList[0].pressureStart / 100000.0,
          tdDive -> mixList[0].pressureEnd / 100000.0);
 
 /* ---- Plot pressure ---- */
 draw_label(mg, buffer, 0.0, 0.875);

 /* ---- Render temperature ---- */
 for (i = 0; i < buffSize; i++) buffer[i]='\0';
 snprintf(buffer,
          buffSize - 1,
          "Temperature: %4.1f C", 
          tdDive -> temperature);
 
 /* ---- Plot temperature ---- */
 draw_label(mg, buffer, 0.0, 0.85);

 /* ---- Init buffer and check for valid surface-interval ---- */
 buffer[0]='\0';

 if (tdDive -> surfaceInterval != C_NAI) {
 
  /* ---- Render the surfaceinterval ----- */ 
  hours = (long) tdDive -> surfaceInterval / 60.0; 
  mins = (long ) tdDive -> surfaceInterval - 60 * hours;
  snprintf(buffer,
           buffSize - 1,
           "Surface time: %2.2ld:%2.2ld", 
           hours,
           mins);
 
 } else {

  /* ---- Render an infinite surface-interval ---- */
  snprintf(buffer,
           buffSize - 1,
           "Surface time: infinity");

 }
 
 /* ---- Surface-interval temperature ---- */
 draw_label(mg, buffer, 0.0, 0.825);

}

void PlotMixes(UDCFDive* tdDive, Multigrapher* mg)
{
 const int buffSize = 1024;
 char buffer[buffSize];
 long i,j;
 double yPos = 0.9;

 /* ---- Check for valid dive-struct ---- */
 if (tdDive == (UDCFDive *) NULL) 
#ifndef NDEBUG
 /* ---- Give an error and exit ---- */
 {
  fprintf(stderr, "Can't plot mix information!\n");
  exit(EXIT_FAILURE);
 }
#else
 /* ---- Simply return if invalid ---- */
 return;
#endif

 /* ---- Iterate over all mixes ---- */
 for (i = 0; i <= tdDive -> mixIndex; i++) {

  /* ---- Render mix-names  ---- */
  for (j = 0; j < buffSize; j++) buffer[j]='\0';
  snprintf(buffer,
           buffSize - 1,
           "MixNo: %ld Name: '%s' Gases (N2/O2/He) (%2.2f/%2.2f/%2.2f)",
           i,
           tdDive -> mixList[i].mixName,
           tdDive -> mixList[i].n2Ratio,
           tdDive -> mixList[i].o2Ratio,
           tdDive -> mixList[i].heRatio);

  /* ---- Draw the mix-info ---- */
  draw_label(mg, buffer, 0.6 * plotWidth, yPos);

  /* ---- Draw the next mix-info below ---- */
  yPos -= 0.025;

 }

}

void PlotGraph(Multigrapher* mg, UDCFDive* dive, bool fill)
{
 const long buffSize = 128;
 char buffer[buffSize];
 double actX = (double) 0.0;
 double actY = (double) 0.0;
 double time, depth; 
 double labelXPos, labelYPos;
 Point point;
 long i;
 long j;

 /* ---- Init the point-struct for plotting ---- */
 point.pendown = fill;
 point.have_x_errorbar = false;
 point.have_y_errorbar = false;
 point.linemode = 3;
 point.use_color = true;
 point.line_width = 0.001;
 if (fill) 
  point.fill_fraction = 0.05;
 else
  point.fill_fraction = -1.0;

 /* ---- Draw the whole profile ---- */
 for (i = 0; i <= dive -> sampleIndex; i++) {

  /* ---- Check for x/y-sample ---- */
  if (dive -> sampleList[i].type == ST_TD) {

   /* ---- Set point to position ---- */
   point.x = dive -> sampleList[i].data.point.x;
   point.y = dive -> sampleList[i].data.point.y;

   /* ---- Store actual position of this sample ---- */
   actX = point.x;
   actY = point.y;

   /* ---- Plot it ---- */
   point.symbol = 1;
   point.symbol_size = 0.05;
   point.linemode = 3;
   plot_point(mg, &point);   

    /* ---- Check for a marker-sample ---- */
  } else if (dive -> sampleList[i].type == ST_MARK) {

   /* ---- Check sample type and set appropriate plot-symbol ---- */
   switch (dive -> sampleList[i].data.mark) {

    case C_ASC: point.symbol = 2;
                break;
    case C_DECO: point.symbol = 3;
                 break;
    case C_SURF: point.symbol = 4;
                 break;
    case C_BREATH: point.symbol = 5;
                   break;
    case C_RBT: point.symbol = 6;
                break;
    default: fprintf(stderr, "Unknown marker type!\n");          
             
   }

   /* ---- Set symbol size and linemode ---- */
   point.symbol_size = 0.04; 
   point.linemode = 5;

   /* ---- If not slow then plot marker at position actX/actY ---- */
   if (dive -> sampleList[i].data.mark == C_SURF) {

    double nextX, nextY;
    long j;

    /* ---- Set the position of the next time/depth-sample to the last one ---- */
    nextX = actX; nextY = actY;

    /* ---- Search for next time/depth-sample ---- */
    j = i + 1;
    while ((dive -> sampleList[j].type != ST_TD) && (j <= dive -> sampleIndex)) j++;

    /* ---- Have we found a time/depth - sample? ---- */
    if (j <= dive -> sampleIndex) {

     /* ---- Set 'nextX' and 'nextY' to position of this sample ---- */
     nextX = dive -> sampleList[j].data.point.x;
     nextY = dive -> sampleList[j].data.point.y;    

    }

    /* ---- Plot the symbol for surface-mode at depth 0.0 ---- */
    point.x = (nextX + actX) / (double) 2.0;
    point.y = 0.0;
    plot_point(mg, &point); 

   } else {

    /* ---- Set symbol position ---- */
    point.x = actX;
    point.y = actY;
  
    /* ---- Plot the symbol ---- */
    plot_point(mg, &point);

   }

  } else if ((dive -> sampleList[i].type == ST_MIX) && (useMixes)) {

   /* ---- Get actual position (time/depth) ---- */
   point.x = actX;
   point.y = actY;

   /* ---- Plot it ---- */
   point.symbol = 8;
   point.symbol_size = 0.03;
   point.linemode = 3;
   plot_point(mg, &point);   
 
   /* ---- Render the label ---- */
   for (j = 0; j < buffSize; j++) buffer[j] = '\0';
   snprintf(buffer,
            buffSize - 1,
            "%ld",
            dive -> sampleList[i].data.mix);

   /* ---- Draw the label for this mix ---- */
   if (!fill) {

    /* ---- Get some dive - data ---- */
    UDCFGetDiveData(dive, &time, &depth);

    /* ---- Calculate X position of label. We have to correct the width/size ---- */  
    /* ---- of font if labelXPos > time                                      ---- */
    labelXPos = actX + plotWidth/plotHeight * 0.5;
    if (labelXPos > time) labelXPos = actX - (plotWidth/plotHeight * 1.00);

    /* ---- Calculate Y position of label. We have to correct the width/size ---- */  
    /* ---- of font if labelYPos > depth                                     ---- */
    labelYPos = actY + 0.5;
    if (labelYPos > depth) labelYPos = actX - 0.9;

    /* ---- Plot the label ---- */
    plot_label(mg, buffer, labelXPos, labelYPos, 0.03);   

   }

  }

 }

}

void PrintRanges(UDCF* profile)
{
 long i;

 /* ---- Print the valid range of groups ----- */
 printf("Valid range of group numbers: 0-%ld\n", profile -> groupIndex);

 /* ---- Not print all ranges for valid dive numbers ----- */
 for (i = 0 ; i <= profile -> groupIndex; i++) {

  /* ---- Print the range for group 'i' ---- */
  printf("Valid range of dive numbers for group %ld: 0-%ld\n", 
         i, 
         profile -> groupList[i].diveIndex);

 } 

 /* ---- Add a simple newline ---- */
 printf("\n");

}
                                                   
void PlotDive(UDCF* profile, 
              const char* format, 
              const char* paper, 
              long groupNo, 
              long diveNo)
{
 Multigrapher* mg = (Multigrapher *) NULL; 
 UDCFDive *tdDive = UDCFConvertDelta2TD(&profile -> groupList[groupNo].diveList[diveNo]);
 double depth = (double) 0.0;
 double time = (double) 0.0;

 /* ---- Constants for multigrapher ---- */
 const char *bg_color = (char *) NULL;     /* color of background, if non-NULL */
 const char *bitmap_size = (char *) NULL;
 const char *max_line_length = (char *) NULL;
 const char *meta_portable = (char *) NULL;
 const char *rotation_angle = (char *) NULL; 
 bool save_screen = false;        /* save screen, i.e. no erase before plot? */

 /* ---- Check for valid dive ---- */
 if (tdDive == (UDCFDive *) NULL) return;

 /* ---- Create new multigrapher object ---- */
 mg = new_multigrapher(format, 
                       bg_color, 
                       bitmap_size, 
                       max_line_length, 
                       meta_portable, 
                       paper, 
                       rotation_angle, 
                       save_screen);

 /* ---- Check whether creation was successful ---- */
 if (mg == (Multigrapher *) NULL) {
  
  /* ---- Give an error and exit ---- */                      
  fprintf(stderr, "Error: couldn't open graphing device\n");
  exit(EXIT_FAILURE);

 }

 /* ---- Calc time and depth of this dive ---- */
 UDCFGetDiveData(&profile -> groupList[groupNo].diveList[diveNo], &time, &depth);

 /* ---- Begin a new graph (push new libplot drawing state onto stack) ---- */
 begin_graph(mg, 1.0, 0.0, 0.0);
 
 /* ---- Set parameters of the plot ---- */
 set_graph_parameters(mg,
                      -0.001,
                      "black",
                       tdDive -> place,
                       NULL,
                       0.07,
                       0.02,
		       AXES_AND_BOX_AND_GRID,
                       0.0, time, 0.0,
                       depth, 0.0, 0.0,
                       false,
                       false,
		       plotWidth, plotHeight,
                       0.0, 0.1,
                       NULL, 0.0525,
                       "Time",
                       NULL, 0.0525,
                       "Depth",
                       false,
                       0,
                       true,
                       0,
                       0,
                       1,
                       1.3,
                       false);                       

 /* ---- Draw all dive-information ---- */
 PlotInfo(tdDive, mg);
 
 /* ---- Draw all mix-information (if needed) ---- */
 if (useMixes) PlotMixes(tdDive, mg);

 /* ---- Plot the dive ---- */
 PlotGraph(mg, tdDive, true);
 PlotGraph(mg, tdDive, false);

 /* ---- End of drawing ---- */
 end_polyline_and_flush(mg);

 /* ---- Draw frame ---- */
 draw_frame_of_graph(mg, false);

 /* ---- End of actual graph ---- */
 end_graph(mg);

 /* ---- Delete the multigrapher-object ---- */
 delete_multigrapher(mg);

 /* ---- Free the newly created dive ---- */
 UDCFFreeDive(tdDive);

}

/* ---- Name of program needed by plotter.c ---- */
const char *progname="udcfplot";

int main(int argc, char* argv[])
{
 int key;
#ifdef HAVE_GETOPTLONG
 int optindex; 
 struct option longargdesc[] = {
                                {"group",   required_argument, NULL, (int)'g'},
                                {"dive",    required_argument, NULL, (int)'d'},
                                {"paper",   required_argument, NULL, (int)'p'},
                                {"format",  required_argument, NULL, (int)'f'},
                                {"ranges",  no_argument,       NULL, (int)'r'},
                                {"help",    no_argument,       NULL, (int)'h'},
                                {"mixes",   no_argument,       NULL, (int)'m'},
                                {"verbose", no_argument,       NULL, (int)'v'},
                                {"input",   required_argument, NULL, (int)'i'},
                                {"output",  required_argument, NULL, (int)'o'},
                                {0,         0,                 NULL, 0       }
                               };
#endif
 UDCF* profile =  (UDCF *) NULL;
 const char* format = formats[0]; /* Default format: postscript */
 const char* paper  = papers[0];  /* Default paper: a4 */
 long groupNo = 0, diveNo = 0;
 bool printRanges = false;

 /* ---- Set 'verbose' to false (No additional messages by default) ---- */
 verbose = false;

 /* ---- By default we show no infos about mixes ---- */
 useMixes = false;

 /* ---- Read all options ---- */
#ifdef HAVE_GETOPTLONG
 while ((key = getopt_long(argc,
                           argv,
                           "hmrvg:d:p:f:o:i:",
                           longargdesc,
                           &optindex)) != EOF) {                                
#else
 while ((key = getopt(argc,
                      argv,
                      "hmrvg:d:p:f:o:i:")) != EOF) {
#endif
 
  /* ---- Check for specific option ---- */
  switch (key) {
   case 'g': /* ---- Zero errno-variable to detect an error after 'strtol' ---- */
             errno = 0;

             /* ---- Convert group-number ---- */
             groupNo = strtol(optarg, (char **) NULL, 10); 

             /* ---- Check if the conversion is ok ---- */
             if (errno != 0) {
              
              /* ---- Give an error and terminate ---- */
              fprintf(stderr, "Wrong group number (%s)!\n", optarg);
              exit(EXIT_FAILURE);             

             } 

             break;

   case 'd': /* ---- Zero errno-variable to detect an error after 'strtol' ---- */
             errno = 0;

             /* ---- Convert group-number ---- */
             diveNo = strtol(optarg, (char **) NULL, 10); 

             /* ---- Check if the conversion is ok ---- */
             if (errno != 0) {
              
              /* ---- Give an error and terminate ---- */
              fprintf(stderr, "Wrong dive number (%s)!\n", optarg);
              exit(EXIT_FAILURE);             

             } 

             break;

   case 'p': /* ---- Check for valid paper switch ---- */
             paper = GetEntry(optarg, papers); 
             break;
   case 'f': /* ---- Check for valid format switch ---- */
             format = GetEntry(optarg, formats); 
             break;
   case 'r':
            /* ---- Set flag that indicates that we have to print the ranges  ---- */
            printRanges = true;
            break;
   case 'h':
            /* ---- Print the help message ---- */
            PrintHelp(argv[0]);
            break;
   case 'v':
            /* ---- Give additional messages ---- */
            verbose = true;
            break;
   case 'm':
            /* ---- Write all infos about uses mixes ---- */
            useMixes = true;
            break;
   case 'i':
            /* ---- Connect 'stdin' to file named 'optarg' ----- */
            ConnectToFile(optarg, O_RDONLY, STDIN_FILENO);
            break;
   case 'o':
            /* ---- Flush stdout ---- */
            fflush(stdout);

            /* ---- Connect 'stdout' to file named 'optarg' ----- */
            ConnectToFile(optarg, O_CREAT|O_WRONLY, STDOUT_FILENO);

            break;
   default:
            printf("Error while parsing options!\n");
            exit(EXIT_FAILURE);
 
  }

 } 

 /* ---- Read data from 'stdin' ---- */
 profile = UDCFRead(stdin);

 /* ---- Check for success ---- */
 if (profile == (UDCF*) NULL) {

  /* ---- Give an error and exit ---- */
  fprintf(stderr, "Invalid UDCF-File!\n");
  exit(EXIT_FAILURE);
 } 

#ifndef NDEBUG
 fprintf(stderr, "GroupNo %ld DiveNo %ld\n", groupNo, diveNo);
#endif
 
 /* ---- Check for valid range of groupNo ---- */
 if (groupNo <= profile -> groupIndex) {

  /* ---- Check for valid range for diveNo ---- */
  if (diveNo <= profile -> groupList[groupNo].diveIndex) {

   /* ---- Check if we have to plot the dive or to print the ranges ---- */
   if (!printRanges) {

    /* ---- Plot the dive ---- */
    PlotDive(profile, format, paper, groupNo, diveNo);
  
   } else {

    /* ---- Print the ranges ---- */
    PrintRanges(profile);

   }

  } else {

   /* ---- Give an error, free data and exit ---- */
   fprintf(stderr, "Invalid dive number (out of range)\n");
   fprintf(stderr, "Use option -r to see valid ranges\n");
   UDCFFree(profile);
   exit(EXIT_FAILURE);

  }

 } else {

  /* ---- Give an error, free data and exit ---- */
  fprintf(stderr, "Invalid group number (out of range)\n");
  UDCFFree(profile);
  exit(EXIT_FAILURE);

 }

 /* ---- Free the data ---- */
 UDCFFree(profile);

 /* ---- Everything is ok ---- */
 return EXIT_SUCCESS;

}
