#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>                                                       
#include <unistd.h>
#include <string.h>

#include <UDCF.h>
#include "bool.h"
#include "fileio.h"
#include "../config.h"

static bool verbose;

void PrintHelp(const char* prgname)
{
#ifdef HAVE_GETOPTLONG
 /* ---- Print help message ---- */
 fprintf(stderr, "Usage: %s [OPTION(s)]\n", prgname);
 if (verbose)
  fprintf(stderr, "$Id: udcfbeautify.c,v 1.5 2000/10/15 18:46:27 streit Exp $\n");
 fprintf(stderr, "-v, --verbose         Give additional information\n");
 fprintf(stderr, "-h, --help            Print a help text\n");
 fprintf(stderr, "-t, --timedepth       Convert all dives to time-depth mode\n");
 fprintf(stderr, "-i, --input filename  Read UDCF-data from 'filename'\n");
 fprintf(stderr, "-o, --output filename Write data to 'filename'\n");
 fprintf(stderr, "\n");
#else
 /* ---- Print help message ---- */
 fprintf(stderr, "Usage: %s [OPTION(s)]\n", prgname);
 if (verbose)
  fprintf(stderr, "$Id: udcfbeautify.c,v 1.5 2000/10/15 18:46:27 streit Exp $\n");
 fprintf(stderr, "-v, Give additional information\n");
 fprintf(stderr, "-h, Print a help text\n");
 fprintf(stderr, "-t, Convert all dives to time-depth mode\n");
 fprintf(stderr, "-i, Read UDCF-data from 'filename'\n");
 fprintf(stderr, "-o, Write data to 'filename'\n");
 fprintf(stderr, "\n");
#endif 

 /* ---- Terminate program successfully ---- */
 exit(EXIT_SUCCESS);
 
}

int main(int argc, char* argv[])
{
 int key;
#ifdef HAVE_GETOPTLONG
 int optindex; 
 struct option longargdesc[] = {
                                {"verbose",   no_argument,       NULL, (int)'v'},
                                {"help",      no_argument,       NULL, (int)'h'},
                                {"timedepth", no_argument,       NULL, (int)'t'},
                                {"input",     required_argument, NULL, (int)'i'},
                                {"output",    required_argument, NULL, (int)'o'},
                                {0,           0,                 NULL, 0       }
                               };
#endif
 UDCF* profile =  (UDCF *) NULL;
 UDCFDive * convertedDive = (UDCFDive *) NULL;
 bool success; 
 bool timeDepth = false;
 long i,j;

 /* ---- Set 'verbose' to false (No additional messages by default) ---- */
 verbose = false;

 /* ---- Read all options ---- */
#ifdef HAVE_GETOPTLONG
 while ((key = getopt_long(argc,
                           argv,
                           "vhti:o:",
                           longargdesc,
                           &optindex)) != EOF) {                                
#else
 while ((key = getopt(argc,
                      argv,
                      "vhti:o:")) != EOF) {
#endif
 
  /* ---- Check for specific option ---- */
  switch (key) {
   case 'h':
            /* ---- Print the help message ---- */
            PrintHelp(argv[0]);
            break;
   case 'v':
            /* ---- Give additional messages ---- */
            verbose = true;
            break;
   case 't':
            /* ---- Set mode to convert all to time-depth mode ---- */ 
            timeDepth = true;
            break;
   case 'i':
            /* ---- Connect 'stdin' to file named 'optarg' ----- */
            ConnectToFile(optarg, O_RDONLY, STDIN_FILENO);
            break;
   case 'o':
            /* ---- Flush stdout ---- */
            fflush(stdout);

            /* ---- Connect 'stdout' to file named 'optarg' ----- */
            ConnectToFile(optarg, O_CREAT|O_WRONLY, STDOUT_FILENO);

            break;
   default:
            printf("Error while parsing options!\n");
            exit(EXIT_FAILURE);
 
  }

 } 

 /* ---- Read data from 'stdin' ---- */
 profile = UDCFRead(stdin);

 /* ---- Check for success ---- */
 if (profile == (UDCF*) NULL) {

  /* ---- Give an error and exit ---- */
  fprintf(stderr, "Invalid UDCF-File!\n");
  exit(EXIT_FAILURE);

 } 

 /* ---- Check if we should convert to time-depth mode ---- */
 if (timeDepth) {

#ifndef NDEBUG
  fprintf(stderr, "Convert to time-depth mode\n");
#endif

  /* ---- Iterate over all groups of dives ---- */
  for (i = 0; i <= profile -> groupIndex; i++) {

   /* ---- Iterate over all groups of dives ---- */
   for (j = 0; j <= profile -> groupList[i].diveIndex; j++) {

    /* ---- Convert the dive ---- */
    convertedDive = UDCFConvertDelta2TD(&profile -> groupList[i].diveList[j]);

    /* ---- Check for success ---- */
    if (convertedDive == (UDCFDive *) NULL) {

     /* ---- Free the profile ---- */
     UDCFFree(profile);
     
     /* ---- Give an error and exit ---- */
     fprintf(stderr, "Can't convert dive %ld!\n", i);
     return EXIT_FAILURE;

    }

    /* ---- Free the old sampleList and copy the converted entries ---- */
    free(profile -> groupList[i].diveList[j].sampleList);
    memcpy(&(profile -> groupList[i].diveList[j]), convertedDive, sizeof(UDCFDive));

   }

  }

 }

 /* ---- Write 'profile' to stdout ---- */
 success = UDCFWrite(stdout, profile);

 /* ---- Free the data ---- */
 UDCFFree(profile);

 /* ---- Check if UDCFWrite was ok ---- */
 if (success) {

  /* ---- Everything is ok ---- */
  return EXIT_SUCCESS;

 } else {

  /* ---- Everything is ok ---- */
  return EXIT_SUCCESS;

 }

}
