#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>                                                       
#include <unistd.h>
#include <errno.h>  
#include <string.h>

#include <UDCF.h>
#include "bool.h"
#include "fileio.h"
#include "../config.h"

static bool verbose;
static char profileFormat[] = "profile_%2.2d%2.2d%2.2d-%2.2d:%2.2d";

void PrintHelp(const char* prgname)
{
#ifdef HAVE_GETOPTLONG
 /* ---- Print help message ---- */
 fprintf(stderr, "Usage: %s [OPTION(s)]\n", prgname);
 if (verbose)
  fprintf(stderr, "$Id: udcfsplit.c,v 1.4 2000/10/15 18:46:27 streit Exp $\n");
 fprintf(stderr, "-v, --verbose         Give additional information\n");
 fprintf(stderr, "-h, --help            Print a help text\n");
 fprintf(stderr, "-i, --input filename  Read UDCF-data from 'filename'\n");
 fprintf(stderr, "\n");
#else
 /* ---- Print help message ---- */
 fprintf(stderr, "Usage: %s [OPTION(s)]\n", prgname);
 if (verbose)
  fprintf(stderr, "$Id: udcfsplit.c,v 1.4 2000/10/15 18:46:27 streit Exp $\n");
 fprintf(stderr, "-v, Give additional information\n");
 fprintf(stderr, "-h, Print a help text\n");
 fprintf(stderr, "-i, Read UDCF-data from 'filename'\n");
 fprintf(stderr, "\n");
#endif 

 /* ---- Terminate program successfully ---- */
 exit(EXIT_SUCCESS);
 
}

bool SplitDive(UDCF* profile, long groupNo, long diveNo)
{
 const int buffLen = 1024;
 UDCF splitProfile;
 UDCFGroup splitGroup[1];
 UDCFDive splitDive[1]; 
 char fileName[buffLen + 1];
 FILE* outFile;
 bool result;

 /* ---- Copy information of dive to split (Note: Dynamic array of samples ---- */
 /* ---- is not copied, we use it only by reference                        ---- */
 splitDive[0] = profile -> groupList[groupNo].diveList[diveNo];

 /* ---- Set dive information in group struct ---- */
 splitGroup[0].diveSize = 0;
 splitGroup[0].diveIndex = 0;
 splitGroup[0].diveList = splitDive;

 /* ---- Copy header information from 'profile' into 'splitProfile' ---- */
 splitProfile = *profile;

 /* ---- Set group information ----- */
 splitProfile.groupSize = 0;
 splitProfile.groupIndex = 0;
 splitProfile.groupList = splitGroup;

 /* ---- Calculate filename ---- */
 snprintf(fileName, 
          buffLen, 
          profileFormat,
          splitDive[0].year,
          splitDive[0].month,
          splitDive[0].day,
          splitDive[0].hour,
          splitDive[0].minute);

 /* ---- Open file named 'fileName' for output ---- */
 outFile = fopen(fileName, "w");

 /* ---- Check the creation ---- */
 if (outFile == (FILE *) NULL) {

  /* ---- Give a warning ---- */
  fprintf(stderr, 
          "Can't open %s for writing the profile! (%s)\n",
          fileName,
          strerror(errno));
 
  /* ---- Simply return with an error ---- */
  return false;
 
 }

 /* ---- Write the file ---- */
 result = UDCFWrite(outFile, &splitProfile);

 /* ---- Close the file ---- */
 fclose(outFile);

 /* ---- Return with result for UDCFWrite ---- */
 return result;

}

int main(int argc, char* argv[])
{
 int key;
#ifdef HAVE_GETOPTLONG
 int optindex; 
 struct option longargdesc[] = {
                                {"verbose",   no_argument,       NULL, (int)'v'},
                                {"help",      no_argument,       NULL, (int)'h'},
                                {"input",     required_argument, NULL, (int)'i'},
                                {0,           0,                 NULL, 0       }
                               };
#endif
 UDCF* profile =  (UDCF *) NULL;
 long groupNo, diveNo;
 bool success = true;

 /* ---- Set 'verbose' to false (No additional messages by default) ---- */
 verbose = false;

 /* ---- Read all options ---- */
#ifdef HAVE_GETOPTLONG
 while ((key = getopt_long(argc,
                           argv,
                           "vhi:",
                           longargdesc,
                           &optindex)) != EOF) {                                
#else
 while ((key = getopt(argc,
                      argv,
                      "vhi:")) != EOF) {
#endif
 
  /* ---- Check for specific option ---- */
  switch (key) {
   case 'h':
            /* ---- Print the help message ---- */
            PrintHelp(argv[0]);
            break;
   case 'v':
            /* ---- Give additional messages ---- */
            verbose = true;
            break;
   case 'i':
            /* ---- Connect 'stdin' to file named 'optarg' ----- */
            ConnectToFile(optarg, O_RDONLY, STDIN_FILENO);
            break;
   default:
            printf("Error while parsing options!\n");
            exit(EXIT_FAILURE);
 
  }

 } 

 /* ---- Read data from 'stdin' ---- */
 profile = UDCFRead(stdin);

 /* ---- Check for success ---- */
 if (profile == (UDCF*) NULL) {

  /* ---- Give an error and exit ---- */
  fprintf(stderr, "Invalid UDCF-File!\n");
  exit(EXIT_FAILURE);

 } 

 /* ---- Iterate over all groups in 'profile' ---- */
 for (groupNo = 0 ; groupNo <= profile -> groupIndex; ++groupNo) {

  /* ---- Iterate over all dive in actual group ---- */
  for (diveNo = 0 ; diveNo <= profile -> groupList[groupNo].diveIndex; ++diveNo) {

   /* ---- Check for verbose mode ---- */
   if (verbose)
    fprintf(stderr, "Split dive %ld in group %ld\n", diveNo, groupNo);

   /* ---- Write actual dive ---- */
   success |= SplitDive(profile, groupNo, diveNo);

  }

 }

 /* ---- Free memory used by 'profile' ---- */
 UDCFFree(profile);

 /* ---- Check wheter all SplitDive calls were ok ---- */
 if (success) {

  /* ---- Everything is ok ---- */
  return EXIT_SUCCESS;
 
 } else {

  /* ---- At least on of the SplitDive calls were wrong ---- */
  return EXIT_FAILURE;

 }

}
